# This is a version of the main.py file found in ../../../server/main.py for testing the plugin locally.
# Use the command `poetry run dev` to run this.
import json
from typing import Optional, Dict, List

import dotenv
import uvicorn
import yaml
from astra_assistants.astra_assistants_manager import AssistantManager
from astra_assistants.tools.structured_code import StructuredProgram, StructuredCodeGenerator, StructuredCodeEditor
from astra_assistants.tools.tool_interface import ToolInterface
from fastapi import FastAPI, HTTPException, Body, Response, Depends, Request, Path
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from fastapi.security.utils import get_authorization_scheme_param
from loguru import logger
import requests
from openai.types.beta.threads.run_submit_tool_outputs_params import ToolOutput
from pydantic import BaseModel, Field

from cass_api.assistant.ddl_tool import MakeDDLTool
from cass_api.assistant.util import RequestBody, SimpleCache, MigrateBody, process_thread, TablesBody
from cass_api.ctags.util import parse_ctags, create_ctags_prompt
from cass_api.fern.util import generate_sdk
from datastore.providers.cassandra_datastore import CassandraDataStore
from starlette.responses import FileResponse

dotenv.load_dotenv()

#datastores = Dict[str, CassandraDataStore]
datastores = {}
ctags_by_token = {}

class OptionalHTTPBearer(HTTPBearer):
    async def __call__(self, request: Request):
        authorization: str = request.headers.get("Authorization")
        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            return None
        return HTTPAuthorizationCredentials(scheme=scheme, credentials=param)

optional_bearer_scheme = OptionalHTTPBearer()

async def verify_token(credentials: Optional[HTTPAuthorizationCredentials] = Depends(optional_bearer_scheme)):
    if credentials is None:
        return None
    if credentials.scheme != "Bearer":
        raise HTTPException(status_code=401, detail="Missing token")
    token = credentials.credentials
    return token

app = FastAPI()

@app.get("/custom-swagger-ui", include_in_schema=False)
async def custom_swagger_ui_html():
    return FileResponse('cass_api/static/custom_swagger_ui.html')

@app.get("/styles.css", include_in_schema=False)
async def styles():
    return FileResponse('cass_api/static/styles.css')

@app.get("/script.js", include_in_schema=False)
async def script():
    return FileResponse('cass_api/static/script.js')


@app.get("/manage.html", include_in_schema=False)
async def manage_html():
    return FileResponse('cass_api/static/manage.html')


@app.get("/openapi_2.json", include_in_schema=False)
async def get_openapi_json(token: Optional[str] = Depends(verify_token)):
    main_schema = app.openapi()
    # TODO potentially re-enable this for vector enabled databases
    #main_schema['components']['schemas']['DeleteRequest']['properties']['ids']['title'] = "DocumentIDs"
    main_schema['paths'] = {path: info for path, info in main_schema['paths'].items() if "/.well-known" not in path}
    if token is None or token not in datastores:
        return Response(content=json.dumps(main_schema))
    datastore = get_datastore_from_cache(token)
    subApp = datastore.getSubApp()
    openapi_schema = subApp.openapi()
    openapi_schema['title'] = "AstraDB Plugin",
    openapi_schema['description'] = "Used for interacting with the user's AstraDB data (AstraDB is an Apache Cassandra based database as a service that contains data organized into keyspaces, tables, and columns) to find answers to questions and retrieve relevant information."
    if 'components' in openapi_schema:
        openapi_schema['components']['schemas'] = openapi_schema['components']['schemas'] | main_schema['components']['schemas']
    for path in main_schema['paths']:
        if "/.well-known" not in path and path not in openapi_schema['paths']:
            openapi_schema['paths'][path] = main_schema['paths'][path]
    return Response(content=json.dumps(openapi_schema))


@app.post(
    "/get_databases",
)
async def get_databases_async(token: str = Depends(verify_token)):
    return get_databases(token)


def get_databases(token: str = Depends(verify_token)):
    # Define the URL
    url = "https://api.astra.datastax.com/v2/databases"

    # Define the headers
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json"
    }

    # Make the GET request
    response = requests.get(url, headers=headers).json()

    # Print the response
    if 'errors' in response:
        # Handle the errors
        errors = response['errors']
        if errors[0]['message']:
            if errors[0]['message'] == 'JWT not valid':
                logger.warning(
                    "Please use the word `token` and your AstraDB token as CASSANDRA_USER and CASSANDRA_PASSWORD respectively instead of client and secret (starting with `ASTRACS:` to allow dynamic astra keyspace creation")
        raise HTTPException(status_code=401, detail="Invalid token")
    else:
        # return only the [].id's
        return [db['id'] for db in response]

class LoginPayload(BaseModel):
    db_id: str

@app.post(
    "/db_login",
    operation_id="auth.db_login",
    openapi_extra={"x-fern-sdk-method-name": "db_login", "x-fern-sdk-group-name": "manage"}
)
async def db_login_async(payload: LoginPayload = Body(...), token: str = Depends(verify_token)) -> Response:
    return db_login(payload, token)

def db_login(payload: LoginPayload = Body(...), token: str = Depends(verify_token)) -> Response:
    global datastores
    datastore = datastores.get(token)
    if datastore is None:
        datastore = CassandraDataStore()
    if datastore.getSubApp():
        subApp = FastAPI()
        datastore.setupEndpoints(subApp)
        datastores[token] = datastore

        content = {"msg": f"Already logged in to {payload.db_id}. Refreshed schema."}
        app.mount(f"",subApp)
        return Response(status_code=409, content=json.dumps(content))
    if payload.db_id == "":
        return Response(status_code=400, content='{"msg": "db_id is required in payload."}')
    datastore.setupSession(token, payload.db_id)
    subApp = FastAPI()
    datastore.setupEndpoints(subApp)
    datastores[token] = datastore
    app.mount(f"",subApp)
    content = {"msg":f"Successfully logged in to {payload.db_id}."}
    return Response(status_code=200, content=json.dumps(content))

PORT = 3333

#origins = [
#    f"http://localhost:{PORT}",
#    "https://chat.openai.com",
#]
#
#app.add_middleware(
#    CORSMiddleware,
#    allow_origins=origins,
#    allow_credentials=True,
#    allow_methods=["*"],
#    allow_headers=["*"],
#)


@app.get("/.well-known/ai-plugin.json")
async def get_manifest():
    file_path = "./local_server/ai-plugin.json"
    simple_headers = {}
    simple_headers["Access-Control-Allow-Private-Network"] = "true"
    return FileResponse(file_path, media_type="text/json", headers=simple_headers)


@app.get("/.well-known/logo.png")
async def get_logo():
    file_path = "./local_server/logo.png"
    return FileResponse(file_path, media_type="text/json")

def get_datastore_from_cache(token) -> CassandraDataStore:
    global datastores
    if token in datastores:
        return datastores[token]
    raise HTTPException(status_code=404, detail="Must login to a database ('/db_login') before calling this endpoint")

@app.get("/.well-known/openapi.yaml")
async def get_openapi_yaml(token: Optional[str] = Depends(verify_token)):
    main_schema = app.openapi()
    # TODO potentially re-enable this for vector enabled databases
    #main_schema['components']['schemas']['DeleteRequest']['properties']['ids']['title'] = "DocumentIDs"
    main_schema['paths'] = {path: info for path, info in main_schema['paths'].items() if "/.well-known" not in path}
    if token is None:
        openapi_yaml = yaml.safe_dump(main_schema)
        return Response(content=openapi_yaml)
    if token not in datastores:
        openapi_yaml = yaml.safe_dump(main_schema)
        return Response(content=openapi_yaml)
    datastore = get_datastore_from_cache(token)
    subApp = datastore.getSubApp()
    openapi_schema = subApp.openapi()
    openapi_schema['title'] = "AstraDB Plugin",
    openapi_schema['description'] = "Used for interacting with the user's AstraDB data (AstraDB is an Apache Cassandra based database as a service that contains data organized into keyspaces, tables, and columns) to find answers to questions and retrieve relevant information."
    openapi_schema['components']['schemas'] = openapi_schema['components']['schemas'] | main_schema['components']['schemas']
    for path in main_schema['paths']:
        if "/.well-known" not in path and path not in openapi_schema['paths']:
            openapi_schema['paths'][path] = main_schema['paths'][path]
    openapi_yaml = yaml.safe_dump(openapi_schema)
    return Response(content=openapi_yaml)

thread_cache = SimpleCache()
ddl_list = []
make_ddl_tool = MakeDDLTool(ddl_list)
instructions="""
The user will give you some sort of object (pydantic, zod, pojo, json, go struct, json schema, etc.) your job is to help them design the appropriate cassandra cql table .

Follow these rules:
 - If timestamps are numeric in sample data use bigint, not timestamp.
 - In your explanation call out primary key / partition key and clustering column when applicable and share valid access patterns.
 - Remember that you must pass partition key when you want to select by clustering column.
 - Do not break down objects into separate components. For nested objects initially just use a TEXT field and suggest it's stored as a json string. Cassandra does not support json types.
 - Do not include WITH clauses in your DDL.
 - The field names should match the object keys as much as possible.
 
Use the make_dll tool to create the DLL statement you include and analyze in your response.

Except for METADATA which is a MAP.
"""
data_modeling_manager = AssistantManager(
    instructions=instructions,
    #model="gpt-3.5-turbo",
    model="openai/gpt-4o-mini",
    #model="groq/llama3-8b-8192",
    name="data modeler",
    tools=[make_ddl_tool]
)

@app.post(
    "/manage/{keyspace_name}/{table_name}/design",
    operation_id="design",
    openapi_extra={"x-fern-ignore": True}
    #openapi_extra={"x-fern-sdk-method-name": "design", "x-fern-sdk-group-name": "manage"}
)
async def design(keyspace_name: str = Path(...), table_name: str = Path(...), query: RequestBody = Body(...), token: str = Depends(verify_token)):
    client = data_modeling_manager.get_client()

    thread = None
    thread_id = thread_cache.get(f"{keyspace_name}-{table_name}")
    if thread_id is None:
        thread = data_modeling_manager.thread
        thread_cache.set(f"{keyspace_name}-{table_name}", thread.id)
        thread_id = thread.id
    else:
        thread = client.beta.threads.retrieve(thread_id=thread_id)

    content = query.content + f"\nNote: use keyspace_name: {keyspace_name} and table_name:{table_name}"
    result: ToolOutput = await data_modeling_manager.run_thread(
        content=content,
        tool=make_ddl_tool,
        thread=thread
    )
    if result['error'] is not None:
        return design(keyspace_name, table_name, result['error'], token)
    ddl = result['output'].to_string(keyspace_name, table_name)
    response = {
        "diagnosis": result['text'],
        "thread_id": thread.id,
        "design_id": result['design_id'],
        "ddl": ddl,
    }

    return response

@app.post(
    "/manage/{keyspace_name}/{table_name}/migrate",
    operation_id="migrate",
    openapi_extra={"x-fern-ignore": True}
    #openapi_extra={"x-fern-sdk-method-name": "design", "x-fern-sdk-group-name": "migrate"}
)
async def migrate(keyspace_name: str = Path(...), table_name: str = Path(...), query: MigrateBody = Body(...), token: str = Depends(verify_token)):
    global datastores
    if token is None:
        raise HTTPException(status_code=401, detail="Missing token header")
    datastore = get_datastore_from_cache(token)
    design_id = query.design_id
    ddl_cache = data_modeling_manager.tools[0].ddl_cache
    for cached in ddl_cache:
        if str(cached['design_id']) == design_id:
            ddl = cached['output'].to_string()
            datastore.client.execute(ddl)
            output = {"msg":f"Created {keyspace_name}.{table_name} using {query} and {ddl}."}

            await db_login(LoginPayload(db_id=datastore.dbid), token)
            response = await get_openapi_yaml(token)
            generate_sdk(response.body.decode("utf-8"), f'{datastore.dbid}.tags')
            return output


programs: List[Dict[str, StructuredProgram]] = []
code_generator = StructuredCodeGenerator(programs)
code_editor = StructuredCodeEditor(programs)


class TrueOrFalsePayload(BaseModel):
    answer: bool = Field(..., description="true or false")

    def to_string(self):
        return "True" if self.answer else "False"

class TrueOrFalseTool(ToolInterface):
    def call(self, result: TrueOrFalsePayload):
        return {"output": result}


true_or_false_tool = TrueOrFalseTool()
tools = [code_generator, code_editor, true_or_false_tool]

code_manager = AssistantManager(
    instructions="use the structured code tool to generate code to help the user.",
    tools=tools,
    model="openai/gpt-4o",
)

@app.get(
    "/manage/ctags",
    operation_id="ctags",
    #openapi_extra={"x-fern-sdk-method-name": "code", "x-fern-sdk-group-name": "manage"}
    openapi_extra={"x-fern-ignore": True}
)
async def get_ctag_tools(token: str = Depends(verify_token)):
    dbid = datastores[token].dbid
    tags_info = parse_ctags(f"./tenant_ctags/{dbid}.tags")
    ctags_prompt = create_ctags_prompt(tags_info)
    ctags_by_token[token] = {"prompt": ctags_prompt, "tags_info": tags_info}
    tables = list(tags_info.keys())
    tables.remove("manage")
    return {"tables": tables}

@app.post(
    "/manage/ctags",
    operation_id="ctags",
    #openapi_extra={"x-fern-sdk-method-name": "code", "x-fern-sdk-group-name": "manage"}
    openapi_extra={"x-fern-ignore": True}
)
async def enable_ctag_tools(payload: TablesBody = Body(...), token: str = Depends(verify_token)):
    tables = payload.tables
    tables.append("manage")
    dbid = datastores[token].dbid
    tags_info = parse_ctags(f"./tenant_ctags/{dbid}.tags")
    tags_info = {table: tags_info[table] for table in tables}
    ctags_prompt = create_ctags_prompt(tags_info)
    code_manager.additional_instructions = ctags_prompt
    return {"success": True}


@app.post(
    "/manage/code",
    operation_id="code",
    #openapi_extra={"x-fern-sdk-method-name": "code", "x-fern-sdk-group-name": "manage"}
    openapi_extra={"x-fern-ignore": True}
)
async def genreate_code(query: RequestBody = Body(...), token: str = Depends(verify_token)):
    output: ToolOutput = await code_manager.run_thread(
        content=query.content,
        tool=code_generator
    )
    result = code_generator.program_cache
    for program in result:
        program['as_string'] = program['output'].to_string()
        program['as_string_no_line_numbers'] = program['output'].to_string(False)
    return result

@app.post(
    "/manage/code/{program_id}",
    operation_id="edit",
    #openapi_extra={"x-fern-sdk-method-name": "code", "x-fern-sdk-group-name": "manage"}
    openapi_extra={"x-fern-ignore": True}
)
async def edit_code(program_id: str = Path(...), query: RequestBody = Body(...), token: str = Depends(verify_token)):
    program_string = None
    for program_obj in code_editor.program_cache:
        if program_obj['program_id'] == program_id:
            program_string = program_obj['output'].to_string()
            break
    if program_string is None:
        return {"error": f"Program id {program_id} not found"}
    output: ToolOutput = await code_manager.run_thread(
        content=f"## Context:\nMake an edit for the program and describe your changes\nprogram_id: {program_id}\nprogram:\n{program_string}\nperform the following changes to the program using the edit tool: {query.content}",
        tool=code_editor
    )
    result = code_editor.program_cache
    for program in result:
        program['as_string'] = program['output'].to_string()
        program['as_string_no_line_numbers'] = program['output'].to_string(False)
        program['text'] = output['text']

    tf_result: ToolOutput = await code_manager.run_thread(
        content=f"This is the output of the edit \n{result[0]['as_string']}\nIs this what you wanted?",
        tool=true_or_false_tool
    )
    print(f"was it what I wanted? {tf_result['output'].answer}")
    #if not tf_result['output'].answer:
    if False:
        print(f"because: {tf_result['text']}")
        print(f"this is not what I wanted, trying again")
        return await edit_code(program_id, query, token)

    return result



def start():
    #uvicorn.run("cass_api.main:app", host="localhost", port=PORT, reload=True, log_level="debug", reload_excludes=["fern_config/*", "tenant_ctags/*"])
    uvicorn.run("cass_api.main:app", host="localhost", port=PORT)

if __name__ == "__main__":
    start()


import asyncio
from typing import List

from astra_assistants.astra_assistants_event_handler import AstraEventHandler
from fastapi import HTTPException
from pydantic import BaseModel


class RequestBody(BaseModel):
    content: str

class TablesBody(BaseModel):
    tables: List[str]

class TableModel(BaseModel):
    partition_key: List[BaseModel]



class MigrateBody(BaseModel):
    design_id: str

class SimpleCache:
    def __init__(self):
        self.cache = {}

    def get(self, key):
        return self.cache.get(key)

    def set(self, key, value):
        self.cache[key] = value

    def clear(self):
        self.cache.clear()


async def process_thread(client, content, thread, assistant_manager):
    assistant = assistant_manager.get_assistant()
    tool = assistant_manager.get_tool()
    event_handler = AstraEventHandler(client)
    event_handler.register_tool(tool)
    try:
        client.beta.threads.messages.create(
            thread_id=thread.id, role="user", content=content
        )
        with client.beta.threads.runs.create_and_stream(
            thread_id=thread.id,
            assistant_id=assistant.id,
            event_handler=event_handler,
            tool_choice=tool.tool_choice_object(),
        ) as stream:
            for part in stream:
                pass
        text = ""
        with event_handler.stream as stream:
            for part in stream.text_deltas:
                text += part

        # TODO: fix this hack for concurrency etc.
        ddl = tool.ddl_cache[len(tool.ddl_cache)-1]

        response = {
            "diagnosis": text,
            "thread_id": thread.id,
            "design_id": ddl['design_id'],
            "ddl": ddl['ddl_statement'],
        }
        print(response)
        return response
    except Exception as e:
        print(e)
        raise HTTPException(status_code=500, detail=str(e))



import json
import logging
import uuid
from typing import List, Dict, Optional

from astra_assistants.tools.tool_interface import ToolInterface
from cassandra.cqltypes import cqltype_to_python
from pydantic import BaseModel, Field, validator

from datastore.providers.cassandra_util import CassandraType

logger = logging.getLogger(__name__)

class Column(BaseModel):
    name: str = Field(..., description="Name of the column")
    type: str = Field(..., description="Cassandra data type of the column. Must be either a base type: \nascii, bigint, blob, boolean, counter, date, decimal, double, duration, float, inet, int, smallint, text, time, timestamp, timeuuid, tinyint, uuid, varchar, varint\nOr a collection where text could be replaced with a base type set<text>, list<text>, map<text, text>\n or finally a vector type which is always float and takes the number of dimensions: vector<float, 1024>")
    @validator('type')
    def validate_type(cls, v):
        try:
            cls._validate_cassandra_type(v)
        except ValueError:
            raise ValueError(f"Unsupported type {v}. Must be a valid Cassandra type.")
        return v

    @staticmethod
    def _validate_cassandra_type(type_str: str):
        python_type = cqltype_to_python(type_str)
        if len(python_type) > 1:
            if python_type[0] == 'set' or python_type[0] == 'list':
                Column._validate_cassandra_type(python_type[1][0])
            elif python_type[0] == 'map':
                Column._validate_cassandra_type(python_type[1][0])
                Column._validate_cassandra_type(python_type[1][1])
            elif python_type[0] == 'vector':
                # No need for further validation; vectors are lists of floats
                return
            else:
                raise ValueError(f"Unsupported collection type {type_str}")
        elif python_type[0] not in {item.value for item in CassandraType}:
            raise ValueError(f"Unsupported type {type_str}")


class DDLModel(BaseModel):
    thoughts: Optional[str] = Field(..., description="The message to be described to the user explaining why the DDL statement is being proposed.")
    keyspace_name: str = Field(..., description="Name of the keyspace")
    table_name: str = Field(..., description="Name of the table")
    columns: List[Column] = Field(..., description="List of columns with their types, this includes partition key and clustering columns")
    partition_key: List[str] = Field(..., description="List of partition key column names, *NOTE ORDER MATTERS*")
    clustering_columns: List[str] = Field(..., description="List of clustering column column names, *NOTE ORDER MATTERS*")

    def to_string(self, keyspace_name: str = None, table_name: str = None):
        if keyspace_name is not None:
            self.keyspace_name = keyspace_name
        if table_name is not None:
            self.table_name = table_name
        ddl_statement = f"CREATE TABLE {self.keyspace_name}.{self.table_name} (\n" + \
            ",\n".join([f"    {column.name} {column.type}" for column in self.columns]) + \
            ",\n    PRIMARY KEY ((" + \
            ", ".join(self.partition_key) + ")"
        if len(self.clustering_columns) > 0:
            ddl_statement += ", " + ", ".join(self.clustering_columns)
        ddl_statement += ")\n);"
        return ddl_statement

class MakeDDLTool(ToolInterface):
    def __init__(self, ddl_list: List):
        self.ddl_cache = ddl_list

    def call(self, arguments: DDLModel) -> str:
        ddl_object = {
            "output": arguments,
            "design_id": uuid.uuid4()
        }
        self.ddl_cache.append(ddl_object)
        return ddl_object


import os
import re
import subprocess
import json

def generate_ctags(directory, output_file):
    """
    Generate ctags for the specified directory and output to the given file.
    """
    if os.path.exists(output_file):
        os.remove(output_file)
    cmd = ['ctags', '-R', '--output-format=json', '--fields=+KnlSa', '--extras=+q', '-f', output_file, directory]
    print(f"about to run command: {cmd}")
    result = subprocess.run(cmd, check=True)
    print(f"result: {result}")


def parse_json_lines_file(file_path):
    tags_info = {'tags': []}
    with open(file_path, 'r') as file:
        for line in file:
            tags_info['tags'].append(json.loads(line))
    return tags_info

def extract_table_name(file_path):
    match = re.search(r'/([^/]+)/client.py', file_path)
    if match:
        return match.group(1)
    return None

def parse_ctags(file_path):
    tags_info = parse_json_lines_file(file_path)


    client_classes_with_retrieve = set()
    for entry in tags_info['tags']:
        if "Client" in entry.get("scope", "") and entry.get("name") == "retrieve":
            client_classes_with_retrieve.add(entry["scope"])
        if "Client" in entry.get("scope", "") and entry.get("name") == "db_login":
            client_classes_with_retrieve.add(entry["scope"])

    relevant_entries = {}
    for entry in tags_info['tags']:
        if entry.get("scope") in client_classes_with_retrieve:
            table = extract_table_name(entry.get("path"))
            if relevant_entries.get(table) is None:
                relevant_entries[table] = []
            relevant_entries[table].append(entry)


    return relevant_entries


def create_ctags_prompt(tags_info):
    prompt = f"""
    You are given a custom SDK with the following ctag information::

    {tags_info}
    
    The SDK usage is similar to the following example (replace keyspace and table with values from the ctag information):
    
    ```python
    import os
    import dotenv
    from src.datastax.client import db

    dotenv.load_dotenv()

    client = db(
        token=os.getenv('ASTRA_DB_APPLICATION_TOKEN'),
    )


    try:
        client.manage.db_login(db_id=os.getenv('ASTRA_DB_ID'))
    except Exception as e:
        if hasattr(e,"status_code") and e.status_code == 409:
            pass
        else:
            print(e)
            raise e

    embedding = [0.37972306566599356, ...]

    client.default_keyspace.document_chunks.create(
        chunk_id="test",
        document_id="test",
        user_id="test",
        workspace_id="test_2",
        chunk_data="a long string",
        embedding=embedding,
    )
    chunk = client.default_keyspace.document_chunks.retrieve(user_id="test", workspace_id="test_2")
    print("chunk" + chunk")

    chunks = client.default_keyspace.document_chunks.search(args={{"embedding": embedding, "user_id": "test", "workspace_id": "test"}})
    print("ann result count: " + len(chunks)")
    ```

    """
    return prompt


import os
import time
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple
from cassandra import ConsistencyLevel, AlreadyExists
import zipfile
import json
import requests
from cassandra.policies import RetryPolicy
from fastapi import Depends, Request, Response, HTTPException
from loguru import logger
from pydantic import BaseModel, create_model

from datastore.datastore import DataStore
from cassandra.cluster import Cluster, NoHostAvailable
from cassandra.auth import PlainTextAuthProvider
from cassandra.query import SimpleStatement, dict_factory, named_tuple_factory, UNSET_VALUE

from datastore.providers.cassandra_util import get_pydantic_type, get_openapi_type
from services.openai import get_embeddings

CASSANDRA_HOST = os.environ.get("CASSANDRA_HOST", "localhost")
CASSANDRA_PORT = int(os.environ.get("CASSANDRA_PORT", 9042))
CASSANDRA_KEYSPACE = os.environ.get("CASSANDRA_KEYSPACE", "default_keyspace")
CASSANDRA_USER = 'token'


class Payload(BaseModel):
    args: Dict[str, Any]

# class that implements the DataStore interface for Cassandra Datastore provider
class CassandraDataStore(DataStore):
    def __init__(self):
        self.app = None
        # no longer create session on init, since we need a session per user
        # self.client = self.create_db_client()
        pass

    def create_db_client(self, token, dbid):
        self.client = CassandraClient(token, dbid)
        return self.client

    def getSubApp(self):
        return self.app

    def setupSession(self, token, dbid):
        self.dbid = dbid
        self.client = self.create_db_client(token, dbid)
        return self.client
    def setupEndpoints(self, app):
        self.app = app
        keyspaces = self.client.get_keyspaces()

        #def create_get_keyspaces_endpoint():
        #    async def endpoint() -> List[str]:
        #        rows = await self.client.get_keyspaces_async()
        #        return rows
        #    return endpoint

        #app.add_api_route(path="/keyspaces", endpoint=create_get_keyspaces_endpoint(), methods=["GET"], operation_id=f"get_keyspaces", openapi_extra={"x-fern-sdk-method-name": "get_keyspaces", "x-fern-sdk-group-name": "manage", "summary": f"Get Keyspaces"})

        for keyspace in keyspaces:

            def create_get_tables_endpoint(keyspace):
                async def endpoint() -> List[str]:
                    rows = await self.client.get_tables_async(keyspace)
                    return rows
                return endpoint

            app.add_api_route(path=f"/{keyspace}/tables", endpoint=create_get_tables_endpoint(keyspace), methods=["GET"], operation_id=f"get_tables", openapi_extra={"x-fern-sdk-method-name": "get_tables", "x-fern-sdk-group-name": ["manage", keyspace], "summary": f"Get Tables"})

            tables = self.client.get_tables(keyspace)

            def create_select_all_endpoint(keyspace: str, table: str, response_model: BaseModel):
                async def endpoint() -> List[response_model]:
                    rows = await self.client.select_all_from_table_async(keyspace, table)
                    return rows

                return endpoint


            def create_get_columns_endpoint(keyspace, table):
                async def endpoint() -> List[Dict[str, Any]]:
                    rows = await self.client.get_columns_async(keyspace, table)
                    return rows

                return endpoint

            def create_select_by_pk_endpoint(keyspace: str, table: str, partition_keys: List[Dict[str, Any]], response_model : BaseModel):
                class PathParams:
                    def __init__(self, keys: Dict[str, Any]):
                        self.keys = keys

                    async def __call__(self, request: Request) -> Dict[str, Any]:
                        return {key['column_name']: request.path_params[key['column_name']] for key in self.keys}

                async def endpoint(args: Dict[str, str] = Depends(PathParams(partition_keys))) -> List[response_model]:
                    rows = await self.client.selectFromTableByPK(keyspace, table, partition_keys, args)
                    return [response_model(**row) for row in rows]
                return endpoint

            def create_select_by_cc_endpoint(keyspace: str, table: str, partitionKeys: List[Dict[str, Any]], clustering_columns: List[Dict[str, Any]], response_model: BaseModel):
                async def endpoint(payload: Payload) -> List[response_model]:
                    rows = await self.client.selectFromTableByPK(keyspace, table, partitionKeys + clustering_columns, payload.args)
                    return rows
                return endpoint


            def create_select_by_index_endpoint(keyspace, table, indexed_columns, vector_indexes, partition_keys, columns, response_model: BaseModel):
                async def endpoint(payload: Payload) -> List[response_model]:
                    try:
                        rows = await self.client.selectFromTableByIndex(
                            keyspace=keyspace,
                            table=table,
                            indexed_columns=indexed_columns,
                            vector_indexes=vector_indexes,
                            partition_keys=partition_keys,
                            columns=columns,
                            args=payload.args
                        )
                        return rows
                    except Exception as e:
                        message_json = {"message": str(e)}
                        raise HTTPException(status_code=500, detail=json.dumps(message_json))

                return endpoint

            def create_insert_endpoint(keyspace: str, table: str, response_model : BaseModel):
                async def endpoint(request_object: response_model) -> Response:
                    await self.client.upsert_table_from_dict_async(keyspace, table, request_object.dict())
                    content = {"success": True}
                    return Response(status_code=200, content=json.dumps(content))
                return endpoint


            for table in tables:
                model_name = ''.join(word.capitalize() for word in keyspace.split('_'))
                model_name += ''.join(word.capitalize() for word in table.split('_'))
                model_name += "Model"

                columns = self.client.get_columns(keyspace, table)
                model_fields: Dict[str, Tuple[Any, Any]] = {}
                for col in columns:
                    if col['kind'] != 'partition_key' and col['kind'] != 'clustering':
                        model_fields[col["column_name"]] = (Optional[get_pydantic_type(col["type"])], None)
                    else:
                        model_fields[col["column_name"]] = (get_pydantic_type(col["type"]), ...)

                ResponseModel = create_model(model_name, **model_fields)

                route_path=f"/{keyspace}/{table}"
                endpoint=create_insert_endpoint(keyspace, table, ResponseModel)
                app.add_api_route(path=route_path, endpoint=endpoint, methods=["POST"], operation_id=f"{keyspace}_{table}_create", openapi_extra={"x-fern-sdk-method-name": "create", "x-fern-sdk-group-name":[keyspace, table], "summary": f"Create: {table}"})


                route_path=f"/{keyspace}/{table}/columns"
                endpoint=create_get_columns_endpoint(keyspace, table)
                app.add_api_route(path=route_path, endpoint=endpoint, methods=["GET"], operation_id=f"{keyspace}_{table}_columns", openapi_extra={"x-fern-sdk-method-name": "columns", "x-fern-sdk-group-name":[keyspace, table], "summary": f"Get Columns: {table}"})

                route_path=f"/{keyspace}/{table}"
                endpoint=create_select_all_endpoint(keyspace, table, ResponseModel)
                app.add_api_route(path=route_path, endpoint=endpoint, methods=["GET"], operation_id=f"{keyspace}_{table}_list", openapi_extra={"x-fern-sdk-method-name": "list", "x-fern-sdk-group-name":[keyspace, table], "summary": f"List: {table}"})

                # filter columns that are partition keys
                partition_keys = [column for column in columns if column['kind'] == 'partition_key']
                endpoint_string = ""
                for column in partition_keys:
                    endpoint_string += "-by-" + column['column_name']
                route_path = f"/{keyspace}/{table}"
                for pk in partition_keys:
                    route_path += "/{" + pk["column_name"] + "}"
                endpoint = create_select_by_pk_endpoint(keyspace, table, partition_keys, ResponseModel)
                app.add_api_route(path=route_path, endpoint=endpoint, methods=["GET"], operation_id=f"{keyspace}_{table}_retrieve", openapi_extra={"x-fern-sdk-method-name": "retrieve", "x-fern-sdk-group-name":[keyspace, table], "summary": f"Retrieve: {table}"})

                clustering_columns = [column for column in columns if column['kind'] == 'clustering']
                cc_so_far = []
                for column in clustering_columns:
                    cc_so_far.append(column)
                    endpoint_string += "-by-" + column['column_name']
                    endpoint_path = f"/{keyspace}/{table}/clustering-column-lookup-{table}" + endpoint_string

                    # Using a function factory to ensure proper closure capture
                    def make_endpoint(keyspace=keyspace, table=table, partition_keys=partition_keys, cc_so_far=list(cc_so_far)):
                        return create_select_by_cc_endpoint(keyspace, table, partition_keys, cc_so_far, ResponseModel)

                    #app.post(endpoint_path)(make_endpoint())
                    # comment out clustering columns for now
                    #app.add_api_route(path=endpoint_path, endpoint=make_endpoint(), methods=["POST"], operation_id=f"{keyspace}_{table}_{endpoint_string}", openapi_extra={"x-fern-sdk-method-name": f"filter{endpoint_string}", "x-fern-sdk-group-name":[keyspace, table], "summary": f"Filter: {table}"})


                indexed_columns = self.client.get_indexes(keyspace, table)
                hasIndex = False
                vector_indexes = []
                # TODO - maybe support indexed collections
                for indexed_column in indexed_columns:
                    hasIndex = True
                    # see if it's a vector column
                    # first get the indexed column's details from columns
                    column_raw = [column for column in columns if column['column_name'] == indexed_column]
                    column = []
                    if len(column_raw) > 0:
                        column = column_raw[0]
                    if len(column) > 0 and 'vector' in column['type']:
                        vector_indexes.append(indexed_column)
                        # create a vector endpoint
                # remove vector columns from indexed columns
                indexed_columns = [column for column in indexed_columns if column not in vector_indexes]
                if hasIndex:
                    endpoint_path = f"/{keyspace}/{table}/index-lookup-{table}"
                    endpoint = create_select_by_index_endpoint(
                        keyspace=keyspace,
                        table=table,
                        indexed_columns=indexed_columns,
                        vector_indexes=vector_indexes,
                        partition_keys=partition_keys,
                        columns=columns,
                        response_model=ResponseModel
                    )
                    try:
                        app.add_api_route(path=endpoint_path, endpoint=endpoint, methods=["POST"], operation_id=f"{keyspace}_{table}_search", openapi_extra={"x-fern-sdk-method-name": f"search", "x-fern-sdk-group-name":[keyspace, table], "summary": f"Search: {table}"})
                    except Exception as e:
                        print(e)
                        detail = {"message": str(e)}
                        raise HTTPException(status_code=500, detail=json.dumps(detail))




        openapi_schema = app.openapi()
        for keyspace in keyspaces:
            tables = self.client.get_tables(keyspace)
            for table in tables:
                columns = self.client.get_columns(keyspace, table)
                # filter columns that are partition keys
                partition_keys = [column for column in columns if column['kind'] == 'partition_key']
                endpoint_string = ""
                pk_path = ""
                parameters = []
                properties = {}
                for column in partition_keys:
                    endpoint_string += "-by-" + column['column_name']
                    properties[column['column_name']] = get_openapi_type(column['type'])
                    properties[column['column_name']]['description'] = column['column_name']
                    pk_path += "/{" + column['column_name'] + "}"
                    # TODO: add examples based on column comments
                    parameters.append({
                        "in": "path",
                        "name": column['column_name'],
                        "required": True,
                        "schema": get_openapi_type(column['type']),
                    })
                openapi_schema['paths'][f"/{keyspace}/{table}" + pk_path]["get"]["operationId"] = f"{keyspace}_{table}_retrieve"
                openapi_schema['paths'][f"/{keyspace}/{table}" + pk_path]["get"]["description"] = f"retrieve {keyspace} {table}"
                openapi_schema['paths'][f"/{keyspace}/{table}" + pk_path]["get"]["parameters"] = parameters
                #body = {
                #    "name": "args",
                #    "required": True,
                #    "content" : {
                #        "application/json": {
                #            "schema": {
                #                "type": "object",
                #                "properties": {
                #                    "args": {
                #                        "type": "object",
                #                        "properties": properties
                #                    }
                #                }
                #            }
                #        }
                #    },
                #    "description": "The partition key columns for keyspace " + keyspace+" and table " + table + " are " + str(partition_keys),
                #}
                #post_operation["requestBody"] = body
                clustering_columns = [column for column in columns if column['kind'] == 'clustering']
                cc_so_far = []
                for column in clustering_columns:
                    cc_so_far.append(column)
                    endpoint_string += "-by-" + column['column_name']
                    properties[column['column_name']] = get_openapi_type(column['type'])
                    properties[column['column_name']]['description'] = column['column_name']
                    # disabling cc query for now
                    #post_operation = openapi_schema['paths'][f"/{keyspace}/{table}/clustering-column-lookup-{table}" + endpoint_string]["post"]
                    body = {
                        "required": True,
                        "content" : {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "args": {
                                            "type": "object",
                                            "properties": deepcopy(properties)
                                        }
                                    }
                                }
                            }
                        },
                        "description": "The partition key columns for keyspace " + keyspace+" and table " + table + " are " +
                                       str(partition_keys) + " and the required clustering columns for this endpoint are " + str(cc_so_far),
                    }
                    #post_operation["requestBody"] = body

                indexed_columns = self.client.get_indexes(keyspace, table)
                if len(indexed_columns) > 0:
                    properties = {}
                    indexed_column_details = []
                    for indexed_column in indexed_columns:
                        # TODO: maybe support indexed collections
                        column_raw = [column for column in columns if column['column_name'] == indexed_column]
                        if len(column_raw)>0:
                            column = column_raw[0]
                            indexed_column_details.append(column)
                            properties[column['column_name']] = get_openapi_type(column['type'])
                            properties[column['column_name']]['description'] = column['column_name']
                    for column in partition_keys:
                        properties[column['column_name']] = get_openapi_type(column['type'])
                        properties[column['column_name']]['description'] = column['column_name']
                        properties[column['column_name']]['nullable'] = True
                    post_operation = openapi_schema['paths'][f"/{keyspace}/{table}/index-lookup-{table}"]["post"]
                    body = {
                        "required": True,
                        "content" : {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "args": {
                                            "type": "object",
                                            "properties": properties
                                        }
                                    }
                                }
                            }
                        },
                        "description": "The indexed columns for keyspace " + keyspace+" and table " + table + " are " + str(indexed_column_details) + " and the partition key columns for keyspace " + keyspace+" and table " + table + " are " + str(partition_keys),
                    }
                    post_operation["requestBody"] = body


        openapi_schema['servers']= [{ "url": "https://www.astraplugin.com" }, { "url": "http://localhost:3333" }]
        if "components" in openapi_schema:
            openapi_schema['components']['securitySchemes'] = {
                "bearerAuth": {
                    "type": "http",
                    "scheme": "bearer"
                }
            }
        # add the endpoints to the openapi schema
        self.app.openapi_schema = openapi_schema


def get_astra_bundle_url(dbid, token):
    # Define the URL
    url = f"https://api.astra.datastax.com/v2/databases/{dbid}/secureBundleURL"

    # Define the headers
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json"
    }

    # Define the payload (if any)
    payload = {}

    # Make the POST request
    response = requests.post(url, headers=headers, data=json.dumps(payload)).json()

    # Print the response
    if 'errors' in response:
        # Handle the errors
        errors = response['errors']
        if errors[0]['message']:
            if errors[0]['message'] == 'JWT not valid':
                logger.warning(
                    "Please use the word `token` and your AstraDB token as CASSANDRA_USER and CASSANDRA_PASSWORD respectively instead of client and secret (starting with `ASTRACS:` to allow dynamic astra keyspace creation")
        return False
    else:
        return response['downloadURL']

def make_keyspace(databaseID, token):
    # Define the URL
    url = f"https://api.astra.datastax.com/v2/databases/{databaseID}/keyspaces/{CASSANDRA_KEYSPACE}"

    # Define the headers
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json"
    }

    # Define the payload (if any)
    payload = {}

    # Make the POST request
    try:
        response = requests.post(url, headers=headers, data=json.dumps(payload)).json()
    except Exception as e:
        logger.debug(f"Failed to create keyspace {CASSANDRA_KEYSPACE} in database {databaseID}: {e}")
        logger.debug(f"Should be because it was hibernated, this should have woken it up. Sleeping 10.")
        time.sleep(10)
        return


    # Print the response
    if 'errors' in response:
        # Handle the errors
        errors = response['errors']
        if errors[0]['message']:
            if errors[0]['message'] == 'JWT not valid':
                logger.warning(
                    "Please use the word `token` and your AstraDB token as CASSANDRA_USER and CASSANDRA_PASSWORD respectively instead of client and secret (starting with `ASTRACS:` to allow dynamic astra keyspace creation")
    return


class VectorRetryPolicy(RetryPolicy):
    def on_read_timeout(self, query, consistency, required_responses,
                        received_responses, data_retrieved, retry_num):
        if retry_num < 3:
            logger.info(f"retrying timeout {retry_num}")
            logger.info(f"query: {query}")
            return RetryPolicy.RETRY, consistency  # return a tuple
        else:
            return RetryPolicy.RETHROW, consistency

    def on_request_error(self, query, consistency, error, retry_num):
        if retry_num < 3:
            logger.info(f"retrying error {retry_num}")
            logger.info(f"query: {query}")
            return RetryPolicy.RETRY, consistency  # return a tuple
        else:
            return RetryPolicy.RETHROW, consistency


    def on_unavailable(self, query, consistency, required_replicas, alive_replicas, retry_num):
        return RetryPolicy.RETHROW, consistency  # return a tuple

    def on_write_timeout(self, query, consistency, write_type,
                         required_responses, received_responses, retry_num):
        return RetryPolicy.RETHROW, consistency  # return a tuple




class CassandraClient():
    def __init__(self, token, dbid) -> None:
        super().__init__()
        self.dbid = dbid
        self.cluster =  None
        try:
            self.connect(token,dbid)
        except NoHostAvailable as e:
            logger.warning(f"No host in the cluster could be contacted: {e}")
            # sleep and retry
            time.sleep(5)
            self.connect(token,dbid)
        except Exception as e:
            logger.warning(f"Exception connecting to cluster: {e}")
            raise e
        # TODO: potentially re-enable document table creation for vector search enabled databases
        #self.create_table(token)

    def connect(self, token, dbid):
        if dbid is not None:
            # connect to Astra
            url = get_astra_bundle_url(dbid, token)
            if url:
                # Download the secure connect bundle and extract it
                r = requests.get(url)
                bundlepath = f'/tmp/{dbid}.zip'
                with open(bundlepath, 'wb') as f:
                    f.write(r.content)
                # Connect to the cluster
                cloud_config = {
                    'secure_connect_bundle': bundlepath
                }
                auth_provider = PlainTextAuthProvider(CASSANDRA_USER, token)
                # TODO - support unhibernating things
                try:
                    self.cluster = Cluster(cloud=cloud_config, auth_provider=auth_provider)
                except Exception as e:
                    make_keyspace(dbid, token)
                    logger.warning(f"DB {dbid} is Hibernated, will attempt to wake it up")
                    return self.connect(token,dbid)

                self.session = self.cluster.connect()
            else:
                #time.sleep(5)
                #return self.connect(token, dbid)
                #print("Failed to get secure bundle URL for token and db")
                raise HTTPException(status_code=400, detail="Failed to establish database connection, please check your astradb token")

    def execute(self, ddl):
        try:
            statement = SimpleStatement(
                ddl
                , consistency_level=ConsistencyLevel.QUORUM
            )
            self.session.execute(statement)
        except Exception as e:
            logger.warning(f"Exception creating table or index: {e}")
            if isinstance(e, AlreadyExists):
                raise HTTPException(status_code=409, detail=f"Failed to create table or index {e}")
            raise HTTPException(status_code=400, detail=f"Failed to create table or index {e}")

    def __del__(self):
        # close the connection when the client is destroyed
        if self.cluster:
            self.cluster.shutdown()


    async def select_all_from_table_async(self, keyspace, table) -> List[Dict[str, Any]]:
        return self.select_all_from_table(keyspace, table)

    def select_all_from_table(self, keyspace, table) -> List[Dict[str, Any]]:
        queryString = f"""SELECT * FROM {keyspace}.{table} limit 10"""
        statement = self.session.prepare(queryString)
        statement.consistency_level = ConsistencyLevel.QUORUM
        self.session.row_factory = dict_factory
        rows = self.session.execute(statement)
        json_rows = [dict(row) for row in rows]
        self.session.row_factory = named_tuple_factory
        return json_rows

    async def selectFromTableByPK(self, keyspace, table, partitionKeys, args) -> List[Dict[str, Any]]:
        queryString = f"""SELECT * FROM {keyspace}.{table} WHERE """
        partitionKeyValues = []
        for column in partitionKeys:
            #TODO support other types (single quotes are for strings)
            queryString += f"{column['column_name']} = ? AND "
            partitionKeyValues.append(args[column['column_name']])
        # remove the last AND
        queryString = queryString[:-4]
        statement = self.session.prepare(queryString)
        statement.consistency_level = ConsistencyLevel.QUORUM
        preparedStatement = statement.bind(partitionKeyValues)
        self.session.row_factory = dict_factory
        rows = self.session.execute(preparedStatement)
        json_rows = [dict(row) for row in rows]
        self.session.row_factory = named_tuple_factory
        return json_rows

    async def upsert_table_from_dict_async(self, keyspace_name: str, table_name : str, obj : Dict):
        return self.upsert_table_from_dict(keyspace_name, table_name, obj)

    def upsert_table_from_dict(self, keyspace_name: str, table_name : str, obj : Dict):
        logger.info(f"going to upsert keyspace {keyspace_name} and table {table_name} using {obj}")
        fields = ', '.join(obj.keys())
        placeholders = ', '.join(['?' for _ in range(len(obj.keys()))])

        values_list = []

        for field in obj.keys():
            value = obj.get(field)
            if value is None:
                formatted_value = UNSET_VALUE
            else:
                formatted_value = value
            values_list.append(formatted_value)

        query_string = f"""insert into {keyspace_name}.{table_name}(
                {fields}
            ) VALUES (
                {placeholders}
            );"""

        statement = self.session.prepare(query_string)
        statement.consistency_level = ConsistencyLevel.QUORUM
        try:
            response = self.session.execute(
                statement,
                tuple(values_list)
            )
        except Exception as e:
            logger.error(f"failed to upsert {table_name}: {obj}")
            raise e

    async def get_tables_async(self, keyspace):
        return self.get_tables(keyspace)

    async def get_keyspaces_async(self):
        return self.get_keyspaces()

    def get_keyspaces(self) -> List[str]:
        queryString = "SELECT DISTINCT keyspace_name FROM system_schema.tables"
        statement = self.session.prepare(queryString)
        statement.consistency_level = ConsistencyLevel.QUORUM
        rows = self.session.execute(statement)
        keyspaces = [row.keyspace_name for row in rows]
        keyspaces.remove("system_auth")
        keyspaces.remove("system_schema")
        keyspaces.remove("system")
        keyspaces.remove("data_endpoint_auth")
        keyspaces.remove("system_traces")
        keyspaces.remove("datastax_sla")
        return keyspaces


    async def get_columns_async(self, keyspace, table) -> List[Dict[str, Any]]:
        return self.get_columns(keyspace, table)

    def get_tables(self, keyspace) -> List[str]:
        queryString = f"""SELECT table_name FROM system_schema.tables WHERE keyspace_name='{keyspace}'"""
        statement = self.session.prepare(queryString)
        statement.consistency_level = ConsistencyLevel.QUORUM
        rows = self.session.execute(statement)
        tables = [row.table_name for row in rows]
        return tables

    def get_indexes(self, keyspace, table):
        queryString = f"""
        SELECT options FROM system_schema.indexes 
        WHERE keyspace_name='{keyspace}' 
        and table_name = '{table}'
        and kind = 'CUSTOM' ALLOW FILTERING;
        """
        statement = self.session.prepare(queryString)
        statement.consistency_level = ConsistencyLevel.QUORUM
        self.session.row_factory = dict_factory
        rows = self.session.execute(statement)
        indexes = [row['options'] for row in rows]
        indexed_columns = []
        for index in indexes:
            options = dict(index)
            # TODO - extract whether it's dot vs cosine for vector types
            # and extract vector type
            if 'StorageAttachedIndex' in options['class_name']:
                indexed_columns.append(options['target'])
        self.session.row_factory = named_tuple_factory
        return indexed_columns

    def get_columns(self, keyspace, table) -> List[Dict[str, Any]]:
        queryString = f"""select column_name, kind, type, position from system_schema."columns" WHERE keyspace_name = '{keyspace}' and table_name = '{table}';"""
        statement = self.session.prepare(queryString)
        statement.consistency_level = ConsistencyLevel.QUORUM
        self.session.row_factory = dict_factory
        rows = self.session.execute(statement)
        json_rows = [dict(row) for row in rows]
        self.session.row_factory = named_tuple_factory
        return json_rows

    async def selectFromTableByIndex(self, keyspace, table, indexed_columns, vector_indexes, partition_keys, columns, args) -> List[Dict[str, Any]]:
        queryString = f"SELECT "
        bind_values = []
        for column in columns:
            if column["column_name"] in vector_indexes and column["column_name"] in args:
                # This will work for all vector tables as long as they use openapi ada-002 to make their embeddings.
                # TODO - support other vector types
                # check if column type is vector of dimension 384 and assume that means e5 embedding. IRL we should encode the embedding algo in the index name
                #if '384' in column['type']:
                #    embeddings.append(get_e5_embeddings([args[column['column_name']]]))
                # otherwise default to davinci-002
                #else:
                # TODO maybe optionally support getting embeddings
                #embeddings.append(get_embeddings([args[column['column_name']]]))
                # TODO maybe support scores one day?
                #queryString += f"similarity_cosine(?, {column['column_name']}) as {column['column_name']}_score, "
                # TODO maybe support optionally not sending back vector values
                queryString += f"{column['column_name']}, "
            else:
                queryString += f"{column['column_name']}, "
        queryString = queryString[:-2]

        has_pk_args = False
        for column in partition_keys:
            if column['column_name'] in args:
                has_pk_args = True

        if len(indexed_columns) > 0 or has_pk_args:
            queryString += f""" FROM {keyspace}.{table} WHERE """
        else:
            queryString += f""" FROM {keyspace}.{table} """

        for column in indexed_columns:
            if args[column]:
                # queryString += f"{column} = '{args[column]}' AND "
                queryString += f"{column} = ? AND "
                bind_values.append(args[column])
        # remove the last AND
        if len(indexed_columns) > 0 and not has_pk_args:
            queryString = queryString[:-4]

        if has_pk_args:
            for column in partition_keys:
                if column['column_name'] in args:
                    queryString += f"{column['column_name']} = ? AND "
                    bind_values.append(args[column['column_name']])
        # remove the last AND
        if has_pk_args:
            queryString = queryString[:-4]

        if len(vector_indexes) > 0:
            queryString += f"ORDER BY "
        for column in vector_indexes:
            if column in args:
                queryString += f"""
                {column} ann of ?
                """
                bind_values.append(args[column])
        # TODO make limit configurable
        queryString += f"LIMIT 5"

        statement = self.session.prepare(queryString)
        statement.retry_policy = VectorRetryPolicy()
        statement.consistency_level = ConsistencyLevel.LOCAL_ONE
        boundStatement = statement.bind(bind_values)
        self.session.row_factory = dict_factory
        rows = self.session.execute(boundStatement, timeout=100)
        json_rows = [dict(row) for row in rows]
        #json_rows = [{k: v for k, v in row.items() if not k in vector_indexes} for row in json_rows]
        self.session.row_factory = named_tuple_factory
        return json_rows
